from mdk.types.builtins import *
from mdk.types.defined import *
from mdk.types.context import StateController
from mdk.types.expressions import ConvertibleExpression, Expression, TupleExpression
from mdk.resources.animation import Animation
from typing import Callable

__all__ = ['AfterImage', 'AfterImageTime', 'AngleAdd', 'AngleDraw', 'AngleMul', 'AngleSet', 'AssertSpecial', 'AttackDist', 'AttackMulSet', 'BindToParent', 'BindToRoot', 'BindToTarget', 'ChangeAnim', 'ChangeAnim2', 'ChangeState', 'ClearClipboard', 'CtrlSet', 'DefenceMulSet', 'DestroySelf', 'EnvColor', 'EnvShake', 'Explod', 'ExplodBindTime', 'ForceFeedback', 'FallEnvShake', 'GameMakeAnim', 'Gravity', 'Helper', 'HitAdd', 'HitBy', 'HitDef', 'HitFallDamage', 'HitFallSet', 'HitFallVel', 'HitOverride', 'HitVelSet', 'LifeAdd', 'LifeSet', 'MakeDust', 'ModifyExplod', 'MoveHitReset', 'NotHitBy', 'Null', 'Offset', 'PalFX', 'ParentVarAdd', 'ParentVarSet', 'Pause', 'PlayerPush', 'PlaySnd', 'PosAdd', 'PosFreeze', 'PosSet', 'PowerAdd', 'PowerSet', 'Projectile', 'RemapPal', 'RemoveExplod', 'ReversalDef', 'ScreenBound', 'SelfState', 'SprPriority', 'StateTypeSet', 'SndPan', 'SuperPause', 'TargetBind', 'TargetDrop', 'TargetFacing', 'TargetLifeAdd', 'TargetPowerAdd', 'TargetState', 'TargetVelAdd', 'TargetVelSet', 'Trans', 'Turn', 'VarAdd', 'VarSet', 'VarRandom', 'VarRangeSet', 'VelAdd', 'VelMul', 'VelSet', 'VictoryQuote', 'Width', 'BGPalFX', 'AllPalFX', 'AppendToClipboard', 'DisplayToClipboard']

def AfterImage(time: ConvertibleExpression | None = None, length: ConvertibleExpression | None = None, palcolor: ConvertibleExpression | None = None, palinvertall: ConvertibleExpression | None = None, palbright: TupleExpression | None = None, palcontrast: TupleExpression | None = None, palpostbright: TupleExpression | None = None, paladd: TupleExpression | None = None, palmul: TupleExpression | None = None, timegap: ConvertibleExpression | None = None, framegap: ConvertibleExpression | None = None, trans: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AfterImageTime(time: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AllPalFX(time: ConvertibleExpression | None = None, add: TupleExpression | None = None, mul: TupleExpression | None = None, sinadd: TupleExpression | None = None, invertall: ConvertibleExpression | None = None, color: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AngleAdd(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AngleDraw(value: ConvertibleExpression | None = None, scale: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AngleMul(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AngleSet(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AppendToClipboard(text: ConvertibleExpression, params: TupleExpression | None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AssertSpecial(flag: ConvertibleExpression, flag2: ConvertibleExpression | None = None, flag3: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AttackDist(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AttackMulSet(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def BGPalFX(time: ConvertibleExpression | None = None, add: TupleExpression | None = None, mul: TupleExpression | None = None, sinadd: TupleExpression | None = None, invertall: ConvertibleExpression | None = None, color: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def BindToParent(time: ConvertibleExpression | None = None, facing: ConvertibleExpression | None = None, pos: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def BindToRoot(time: ConvertibleExpression | None = None, facing: ConvertibleExpression | None = None, pos: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def BindToTarget(time: ConvertibleExpression | None = None, id: ConvertibleExpression | None = None, pos: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ChangeAnim(value: ConvertibleExpression | Animation, elem: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ChangeAnim2(value: ConvertibleExpression | Animation, elem: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ChangeState(value: ConvertibleExpression | str | int | Callable[..., None | StateController], ctrl: ConvertibleExpression | None = None, anim: ConvertibleExpression | Animation | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ClearClipboard(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def CtrlSet(ctrl: ConvertibleExpression | None = None, value: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def DefenceMulSet(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def DestroySelf(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def DisplayToClipboard(text: ConvertibleExpression, params: TupleExpression | None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def EnvColor(value: TupleExpression | None = None, time: ConvertibleExpression | None = None, under: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def EnvShake(time: ConvertibleExpression, freq: ConvertibleExpression | None = None, ampl: ConvertibleExpression | None = None, phase: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Explod(anim: ConvertibleExpression | Animation, id: ConvertibleExpression | None = None, pos: TupleExpression | None = None, postype: ConvertibleExpression | None = None, space: ConvertibleExpression | None = None, facing: ConvertibleExpression | None = None, vfacing: ConvertibleExpression | None = None, bindtime: ConvertibleExpression | None = None, vel: TupleExpression | None = None, accel: TupleExpression | None = None, random: TupleExpression | None = None, removetime: ConvertibleExpression | None = None, supermove: ConvertibleExpression | None = None, supermovetime: ConvertibleExpression | None = None, pausemovetime: ConvertibleExpression | None = None, scale: TupleExpression | None = None, sprpriority: ConvertibleExpression | None = None, ontop: ConvertibleExpression | None = None, shadow: ConvertibleExpression | None = None, ownpal: ConvertibleExpression | None = None, removeongethit: ConvertibleExpression | None = None, trans: ConvertibleExpression | None = None, angle: ConvertibleExpression | None = None, alpha: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ExplodBindTime(id: ConvertibleExpression | None = None, time: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ForceFeedback(waveform: ConvertibleExpression | None = None, time: ConvertibleExpression | None = None, freq: TupleExpression | None = None, ampl: TupleExpression | None = None, self: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def FallEnvShake(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def GameMakeAnim(value: ConvertibleExpression | None = None, under: ConvertibleExpression | None = None, pos: TupleExpression | None = None, random: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Gravity(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Helper(helpertype: ConvertibleExpression | None = None, name: ConvertibleExpression | None = None, id: ConvertibleExpression | None = None, pos: TupleExpression | None = None, postype: ConvertibleExpression | None = None, facing: ConvertibleExpression | None = None, stateno: ConvertibleExpression | str | int | Callable[..., None | StateController] | None = None, keyctrl: ConvertibleExpression | None = None, ownpal: ConvertibleExpression | None = None, supermovetime: ConvertibleExpression | None = None, pausemovetime: ConvertibleExpression | None = None, size_xscale: ConvertibleExpression | None = None, size_yscale: ConvertibleExpression | None = None, size_ground_back: ConvertibleExpression | None = None, size_ground_front: ConvertibleExpression | None = None, size_air_back: ConvertibleExpression | None = None, size_air_front: ConvertibleExpression | None = None, size_height: ConvertibleExpression | None = None, size_proj_doscale: ConvertibleExpression | None = None, size_head_pos: TupleExpression | None = None, size_mid_pos: TupleExpression | None = None, size_shadowoffset: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def HitAdd(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def HitBy(value: TupleExpression | None = None, value2: TupleExpression | None = None, time: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def HitDef(attr: TupleExpression, hitflag: ConvertibleExpression | None = None, guardflag: ConvertibleExpression | None = None, affectteam: ConvertibleExpression | None = None, animtype: ConvertibleExpression | None = None, air_animtype: ConvertibleExpression | None = None, fall_animtype: ConvertibleExpression | None = None, priority: TupleExpression | None = None, damage: ConvertibleExpression | None = None, pausetime: TupleExpression | None = None, guard_pausetime: TupleExpression | None = None, sparkno: ConvertibleExpression | None = None, guard_sparkno: ConvertibleExpression | None = None, sparkxy: TupleExpression | None = None, hitsound: TupleExpression | None = None, guardsound: TupleExpression | None = None, ground_type: ConvertibleExpression | None = None, air_type: ConvertibleExpression | None = None, ground_slidetime: ConvertibleExpression | None = None, guard_slidetime: ConvertibleExpression | None = None, ground_hittime: ConvertibleExpression | None = None, guard_hittime: ConvertibleExpression | None = None, air_hittime: ConvertibleExpression | None = None, guard_ctrltime: ConvertibleExpression | None = None, guard_dist: ConvertibleExpression | None = None, yaccel: ConvertibleExpression | None = None, ground_velocity: TupleExpression | None = None, guard_velocity: ConvertibleExpression | None = None, air_velocity: TupleExpression | None = None, airguard_velocity: TupleExpression | None = None, ground_cornerpush_veloff: ConvertibleExpression | None = None, air_cornerpush_veloff: ConvertibleExpression | None = None, down_cornerpush_veloff: ConvertibleExpression | None = None, guard_cornerpush_veloff: ConvertibleExpression | None = None, airguard_cornerpush_veloff: ConvertibleExpression | None = None, airguard_ctrltime: ConvertibleExpression | None = None, air_juggle: ConvertibleExpression | None = None, mindist: TupleExpression | None = None, maxdist: TupleExpression | None = None, snap: TupleExpression | None = None, p1sprpriority: ConvertibleExpression | None = None, p2sprpriority: ConvertibleExpression | None = None, p1facing: ConvertibleExpression | None = None, p1getp2facing: ConvertibleExpression | None = None, p2facing: ConvertibleExpression | None = None, p1stateno: ConvertibleExpression | str | int | Callable[..., None | StateController] | None = None, p2stateno: ConvertibleExpression | str | int | Callable[..., None | StateController] | None = None, p2getp1state: ConvertibleExpression | None = None, forcestand: ConvertibleExpression | None = None, fall: ConvertibleExpression | None = None, fall_xvelocity: ConvertibleExpression | None = None, fall_yvelocity: ConvertibleExpression | None = None, fall_recover: ConvertibleExpression | None = None, fall_recovertime: ConvertibleExpression | None = None, fall_damage: ConvertibleExpression | None = None, air_fall: ConvertibleExpression | None = None, forcenofall: ConvertibleExpression | None = None, down_velocity: TupleExpression | None = None, down_hittime: ConvertibleExpression | None = None, down_bounce: ConvertibleExpression | None = None, chainid: ConvertibleExpression | None = None, nochainid: TupleExpression | None = None, hitonce: ConvertibleExpression | None = None, kill: ConvertibleExpression | None = None, guard_kill: ConvertibleExpression | None = None, fall_kill: ConvertibleExpression | None = None, numhits: ConvertibleExpression | None = None, getpower: TupleExpression | None = None, givepower: TupleExpression | None = None, palfx_time: ConvertibleExpression | None = None, palfx_mul: TupleExpression | None = None, palfx_add: TupleExpression | None = None, envshake_time: ConvertibleExpression | None = None, envshake_freq: ConvertibleExpression | None = None, envshake_ampl: ConvertibleExpression | None = None, envshake_phase: ConvertibleExpression | None = None, fall_envshake_time: ConvertibleExpression | None = None, fall_envshake_freq: ConvertibleExpression | None = None, fall_envshake_ampl: ConvertibleExpression | None = None, fall_envshake_phase: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def HitFallDamage(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def HitFallSet(value: ConvertibleExpression | None = None, xvel: ConvertibleExpression | None = None, yvel: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def HitFallVel(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def HitOverride(attr: TupleExpression, stateno: ConvertibleExpression | str | int | Callable[..., None | StateController] | None = None, slot: ConvertibleExpression | None = None, time: ConvertibleExpression | None = None, forceair: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def HitVelSet(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def LifeAdd(value: ConvertibleExpression, kill: ConvertibleExpression | None = None, absolute: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def LifeSet(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def MakeDust(pos: TupleExpression | None = None, pos2: TupleExpression | None = None, spacing: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ModifyExplod(id: ConvertibleExpression, pos: TupleExpression | None = None, postype: ConvertibleExpression | None = None, facing: ConvertibleExpression | None = None, vfacing: ConvertibleExpression | None = None, bindtime: ConvertibleExpression | None = None, vel: TupleExpression | None = None, accel: TupleExpression | None = None, random: TupleExpression | None = None, removetime: ConvertibleExpression | None = None, supermove: ConvertibleExpression | None = None, supermovetime: ConvertibleExpression | None = None, pausemovetime: ConvertibleExpression | None = None, scale: TupleExpression | None = None, sprpriority: ConvertibleExpression | None = None, ontop: ConvertibleExpression | None = None, shadow: ConvertibleExpression | None = None, ownpal: ConvertibleExpression | None = None, removeongethit: ConvertibleExpression | None = None, trans: ConvertibleExpression | None = None, alpha: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def MoveHitReset(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def NotHitBy(value: TupleExpression | None = None, value2: TupleExpression | None = None, time: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Null(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Offset(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def PalFX(time: ConvertibleExpression | None = None, add: TupleExpression | None = None, mul: TupleExpression | None = None, sinadd: TupleExpression | None = None, invertall: ConvertibleExpression | None = None, color: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ParentVarAdd(var: ConvertibleExpression | None = None, value: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ParentVarSet(var: ConvertibleExpression | None = None, value: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Pause(time: ConvertibleExpression, endcmdbuftime: ConvertibleExpression | None = None, movetime: ConvertibleExpression | None = None, pausebg: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def PlayerPush(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def PlaySnd(value: TupleExpression, volumescale: ConvertibleExpression | None = None, channel: ConvertibleExpression | None = None, lowpriority: ConvertibleExpression | None = None, freqmul: ConvertibleExpression | None = None, loop: ConvertibleExpression | None = None, pan: ConvertibleExpression | None = None, abspan: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def PosAdd(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def PosFreeze(value: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def PosSet(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def PowerAdd(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def PowerSet(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Projectile(attr: TupleExpression, projid: ConvertibleExpression | None = None, projanim: ConvertibleExpression | Animation | None = None, projhitanim: ConvertibleExpression | Animation | None = None, projremanim: ConvertibleExpression | Animation | None = None, projscale: TupleExpression | None = None, projremove: ConvertibleExpression | None = None, projremovetime: ConvertibleExpression | None = None, velocity: TupleExpression | None = None, remvelocity: TupleExpression | None = None, accel: TupleExpression | None = None, velmul: TupleExpression | None = None, projhits: ConvertibleExpression | None = None, projmisstime: ConvertibleExpression | None = None, projpriority: ConvertibleExpression | None = None, projedgebound: ConvertibleExpression | None = None, projstagebound: ConvertibleExpression | None = None, projheightbound: TupleExpression | None = None, offset: TupleExpression | None = None, postype: ConvertibleExpression | None = None, projshadow: ConvertibleExpression | None = None, supermovetime: ConvertibleExpression | None = None, pausemovetime: ConvertibleExpression | None = None, afterimage_time: ConvertibleExpression | None = None, afterimage_length: ConvertibleExpression | None = None, afterimage_palcolor: ConvertibleExpression | None = None, afterimage_palinvertall: ConvertibleExpression | None = None, afterimage_palbright: TupleExpression | None = None, afterimage_palcontrast: TupleExpression | None = None, afterimage_palpostbright: TupleExpression | None = None, afterimage_paladd: TupleExpression | None = None, afterimage_palmul: TupleExpression | None = None, afterimage_timegap: ConvertibleExpression | None = None, afterimage_framegap: ConvertibleExpression | None = None, afterimage_trans: ConvertibleExpression | None = None, hitflag: ConvertibleExpression | None = None, guardflag: ConvertibleExpression | None = None, affectteam: ConvertibleExpression | None = None, animtype: ConvertibleExpression | None = None, air_animtype: ConvertibleExpression | None = None, fall_animtype: ConvertibleExpression | None = None, priority: TupleExpression | None = None, damage: ConvertibleExpression | None = None, pausetime: ConvertibleExpression | None = None, guard_pausetime: ConvertibleExpression | None = None, sparkno: ConvertibleExpression | None = None, guard_sparkno: ConvertibleExpression | None = None, sparkxy: TupleExpression | None = None, hitsound: TupleExpression | None = None, guardsound: TupleExpression | None = None, ground_type: ConvertibleExpression | None = None, air_type: ConvertibleExpression | None = None, ground_slidetime: ConvertibleExpression | None = None, guard_slidetime: ConvertibleExpression | None = None, ground_hittime: ConvertibleExpression | None = None, guard_hittime: ConvertibleExpression | None = None, air_hittime: ConvertibleExpression | None = None, guard_ctrltime: ConvertibleExpression | None = None, guard_dist: ConvertibleExpression | None = None, yaccel: ConvertibleExpression | None = None, ground_velocity: TupleExpression | None = None, guard_velocity: TupleExpression | None = None, air_velocity: TupleExpression | None = None, airguard_velocity: TupleExpression | None = None, ground_cornerpush_veloff: ConvertibleExpression | None = None, air_cornerpush_veloff: ConvertibleExpression | None = None, down_cornerpush_veloff: ConvertibleExpression | None = None, guard_cornerpush_veloff: ConvertibleExpression | None = None, airguard_cornerpush_veloff: ConvertibleExpression | None = None, airguard_ctrltime: ConvertibleExpression | None = None, air_juggle: ConvertibleExpression | None = None, mindist: TupleExpression | None = None, maxdist: TupleExpression | None = None, snap: TupleExpression | None = None, p1sprpriority: ConvertibleExpression | None = None, p2sprpriority: ConvertibleExpression | None = None, p1facing: ConvertibleExpression | None = None, p1getp2facing: ConvertibleExpression | None = None, p2facing: ConvertibleExpression | None = None, p1stateno: ConvertibleExpression | str | int | Callable[..., None | StateController] | None = None, p2stateno: ConvertibleExpression | str | int | Callable[..., None | StateController] | None = None, p2getp1state: ConvertibleExpression | None = None, forcestand: ConvertibleExpression | None = None, fall: ConvertibleExpression | None = None, fall_xvelocity: ConvertibleExpression | None = None, fall_yvelocity: ConvertibleExpression | None = None, fall_recover: ConvertibleExpression | None = None, fall_recovertime: ConvertibleExpression | None = None, fall_damage: ConvertibleExpression | None = None, air_fall: ConvertibleExpression | None = None, forcenofall: ConvertibleExpression | None = None, down_velocity: TupleExpression | None = None, down_hittime: ConvertibleExpression | None = None, down_bounce: ConvertibleExpression | None = None, chainid: ConvertibleExpression | None = None, nochainid: TupleExpression | None = None, hitonce: ConvertibleExpression | None = None, kill: ConvertibleExpression | None = None, guard_kill: ConvertibleExpression | None = None, fall_kill: ConvertibleExpression | None = None, numhits: ConvertibleExpression | None = None, getpower: TupleExpression | None = None, givepower: TupleExpression | None = None, palfx_time: ConvertibleExpression | None = None, palfx_mul: TupleExpression | None = None, palfx_add: TupleExpression | None = None, envshake_time: ConvertibleExpression | None = None, envshake_freq: ConvertibleExpression | None = None, envshake_ampl: ConvertibleExpression | None = None, envshake_phase: ConvertibleExpression | None = None, fall_envshake_time: ConvertibleExpression | None = None, fall_envshake_freq: ConvertibleExpression | None = None, fall_envshake_ampl: ConvertibleExpression | None = None, fall_envshake_phase: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def RemapPal(source: TupleExpression, dest: TupleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def RemoveExplod(id: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ReversalDef(reversal_attr: TupleExpression, hitflag: ConvertibleExpression | None = None, pausetime: TupleExpression | None = None, sparkno: ConvertibleExpression | None = None, guard_sparkno: ConvertibleExpression | None = None, hitsound: TupleExpression | None = None, guardsound: TupleExpression | None = None, p1stateno: ConvertibleExpression | str | int | Callable[..., None | StateController] | None = None, p2stateno: ConvertibleExpression | str | int | Callable[..., None | StateController] | None = None, p1sprpriority: ConvertibleExpression | None = None, p2sprpriority: ConvertibleExpression | None = None, sparkxy: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ScreenBound(value: ConvertibleExpression | None = None, movecamera: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def SelfState(value: ConvertibleExpression, ctrl: ConvertibleExpression | None = None, anim: ConvertibleExpression | Animation | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def SprPriority(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def StateTypeSet(statetype: ConvertibleExpression | None = None, movetype: ConvertibleExpression | None = None, physics: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def SndPan(channel: ConvertibleExpression, pan: ConvertibleExpression, abspan: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def SuperPause(time: ConvertibleExpression | None = None, anim: ConvertibleExpression | None = None, sound: TupleExpression | None = None, pos: TupleExpression | None = None, darken: ConvertibleExpression | None = None, p2defmul: ConvertibleExpression | None = None, poweradd: ConvertibleExpression | None = None, unhittable: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def TargetBind(time: ConvertibleExpression | None = None, id: ConvertibleExpression | None = None, pos: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def TargetDrop(excludeid: ConvertibleExpression | None = None, keepone: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def TargetFacing(value: ConvertibleExpression, id: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def TargetLifeAdd(value: ConvertibleExpression, id: ConvertibleExpression | None = None, kill: ConvertibleExpression | None = None, absolute: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def TargetPowerAdd(value: ConvertibleExpression, id: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def TargetState(value: ConvertibleExpression | str | int | Callable[..., None | StateController], id: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def TargetVelAdd(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, id: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def TargetVelSet(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, id: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Trans(trans: ConvertibleExpression, alpha: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Turn(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def VarAdd(var: ConvertibleExpression | None = None, value: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def VarSet(var: ConvertibleExpression | None = None, value: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def VarRandom(v: ConvertibleExpression, range: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def VarRangeSet(value: ConvertibleExpression, fvalue: ConvertibleExpression, first: ConvertibleExpression | None = None, last: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def VelAdd(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def VelMul(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def VelSet(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def VictoryQuote(value: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Width(edge: TupleExpression | None = None, player: TupleExpression | None = None, value: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...



import traceback
import functools
import copy
from typing import Optional

from mdk.types.expressions import Expression, ConvertibleExpression
from mdk.types.specifier import TypeSpecifier
from mdk.types.context import CompilerContext, ParameterDefinition, StateController, StateScope
from mdk.types.builtins import IntType, FloatType, ShortType, ByteType, BoolType

from mdk.utils.shared import generate_random_string, convert, format_bool

## a special type of Expression which represents a variable access.
## generally speaking this is just treated differently so that we can
## detect variable initialization and scope in the state/template code.
class VariableExpression(Expression):
    def __init__(self, type: TypeSpecifier, scope: Optional[StateScope] = None, system: bool = False):
        self.type = type
        self.exprn = ""
        self.system = system

        ## in order to determine a name for this variable, we need to walk through the backtrace
        ## and find calling code which originates from a statedef or template function, or
        ## from the top level of a module (other than this one).
        context = CompilerContext.instance()
        traceback.extract_tb(None)
        for frame in traceback.extract_stack():
            function_name = frame.name
            if function_name in context.statedefs and frame.line != None and len(frame.line.split("=")) == 2:
                ## means this is a line from a statedef or template function, so we should check the assignment
                self.exprn = frame.line.split("=")[0].strip()
                if next(filter(lambda k: k.name == self.exprn, context.statedefs[function_name].locals), None) != None:
                    raise Exception(f"Attempted to create 2 local variables with the same name {self.exprn} in state definition {function_name}.")
                context.statedefs[function_name].locals.append(ParameterDefinition(self.type, self.exprn))
                break
            elif function_name in context.templates and frame.line != None and len(frame.line.split("=")) == 2:
                ## means this is a line from a statedef or template function, so we should check the assignment
                self.exprn = frame.line.split("=")[0].strip()
                if next(filter(lambda k: k.name == self.exprn, context.templates[function_name].locals), None) != None:
                    raise Exception(f"Attempted to create 2 local variables with the same name {self.exprn} in template definition {function_name}.")
                context.templates[function_name].locals.append(ParameterDefinition(self.type, self.exprn))
            elif context.current_state == None and context.current_template == None and function_name == "<module>" and frame.line != None and len(frame.line.split("=")) >= 2:
                ## means this is a line from a global scope
                self.exprn = frame.line.split("=")[0].strip()
                if next(filter(lambda k: k.name == self.exprn, context.globals), None) != None:
                    raise Exception(f"Attempted to create 2 global variables with the same name {self.exprn}.")
                context.globals.append(ParameterDefinition(self.type, self.exprn, scope = scope, is_system = self.system))
        
        ## if it could not be found, print a warning and assign a variable name.
        if context.current_state != None and self.exprn == "":
            self.exprn = f"_anon_{generate_random_string(8)}"
            print(f"Warning: Could not automatically identify the name of a variable in {context.current_state.fn.__name__}, assigning anonymous name {self.exprn}.")
        elif context.current_template != None and self.exprn == "":
            self.exprn = f"_anon_{generate_random_string(8)}"
            print(f"Warning: Could not automatically identify the name of a variable in {context.current_template.fn.__name__}, assigning anonymous name {self.exprn}.")
        elif context.current_template == None and context.current_state == None and self.exprn == "":
            self.exprn = f"_anon_{generate_random_string(8)}"
            print(f"Warning: Could not automatically identify the name of a variable in global state, assigning anonymous name {self.exprn}.")

    def make_expression(self, exprn: str):
        return Expression(exprn, self.type)
    
    def set(self, val: ConvertibleExpression):
        context = CompilerContext.instance()
        new_value = convert(val)
        new_controller = StateController()
        new_controller.type = "VarSet"
        new_controller.params = {}
        new_controller.params[self.exprn] = new_value
        if len(context.trigger_stack) != 0 or len(context.if_stack) != 0:
            new_controller.triggers = copy.deepcopy(context.trigger_stack)
            if len(context.if_stack) != 0:
                new_controller.triggers.append(Expression(f"mdk_internalTrigger = {context.if_stack[-1]}", BoolType))
            if len(context.check_stack) != 0:
                new_controller.triggers = copy.deepcopy(context.check_stack) + new_controller.triggers
                context.check_stack = []
        else:
            new_controller.triggers = [format_bool(True)]
        ## i believe this should always be right, since it flows from <callsite> -> .set
        callsite = traceback.extract_stack()[-2]
        new_controller.location = (callsite.filename, callsite.lineno if callsite.lineno != None else 0)

        if context.default_state[0] != None: new_controller.params["ignorehitpause"] = context.default_state[0]
        if context.default_state[1] != None: new_controller.params["persistent"] = context.default_state[1]

        if context.current_state != None:
            context.current_state.controllers.append(new_controller)
        elif context.current_template != None:
            context.current_template.controllers.append(new_controller)

    def add(self, val: ConvertibleExpression):
        context = CompilerContext.instance()
        new_value = convert(val)
        new_controller = StateController()
        new_controller.type = "VarAdd"
        new_controller.params = {}
        new_controller.params[self.exprn] = new_value
        if len(context.trigger_stack) != 0 or len(context.if_stack) != 0:
            new_controller.triggers = copy.deepcopy(context.trigger_stack)
            if len(context.if_stack) != 0:
                new_controller.triggers.append(Expression(f"mdk_internalTrigger = {context.if_stack[-1]}", BoolType))
            if len(context.check_stack) != 0:
                new_controller.triggers = copy.deepcopy(context.check_stack) + new_controller.triggers
                context.check_stack = []
        else:
            new_controller.triggers = [format_bool(True)]
        ## i believe this should always be right, since it flows from <callsite> -> .set
        callsite = traceback.extract_stack()[-2]
        new_controller.location = (callsite.filename, callsite.lineno if callsite.lineno != None else 0)

        if context.default_state[0] != None: new_controller.params["ignorehitpause"] = context.default_state[0]
        if context.default_state[1] != None: new_controller.params["persistent"] = context.default_state[1]

        if context.current_state != None:
            context.current_state.controllers.append(new_controller)
        elif context.current_template != None:
            context.current_template.controllers.append(new_controller)
    
## these are helpers for creating variables from commonly-used built-in types.
IntVar = functools.partial(VariableExpression, type = IntType)
ShortVar = functools.partial(VariableExpression, type = ShortType)
ByteVar = functools.partial(VariableExpression, type = ByteType)
BoolVar = functools.partial(VariableExpression, type = BoolType)
FloatVar = functools.partial(VariableExpression, type = FloatType)

__all__ = ["VariableExpression", "IntVar", "FloatVar", "ShortVar", "ByteVar", "BoolVar"]


from dataclasses import dataclass
from enum import Enum, Flag
from typing import Union, Optional

from mdk.types.specifier import TypeCategory, TypeSpecifier
from mdk.types.builtins import *
from mdk.types.expressions import Expression
from mdk.types.context import CompilerContext
from mdk.types.enums import *

## represents a structure member, mapping its field name to a type.
@dataclass
class StructureMember:
    name: str
    type: TypeSpecifier

class StructureType(TypeSpecifier):
    members: list[StructureMember]
    def __init__(self, name: str, members: list[StructureMember], register: bool = True, library: str | None = None):
        self.name = name
        self.category = TypeCategory.STRUCTURE
        self.members = members
        self.register = register

        context = CompilerContext.instance()
        if self.name in context.typedefs:
            raise Exception(f"A type with name {self.name} was already registered.")
        context.typedefs[self.name] = self

        self.library = library

class EnumType(TypeSpecifier):
    members: list[str]
    inner_type: Optional[type[Enum]]
    def __init__(self, name: str, members: Union[list[str], type[Enum]], register: bool = True, library: str | None = None):
        self.name = name
        self.category = TypeCategory.ENUM
        if isinstance(members, list):
            self.members = members
            self.inner_type = None
        else:
            self.members = members._member_names_
            self.inner_type = members
        self.register = register

        context = CompilerContext.instance()
        if self.name in context.typedefs:
            raise Exception(f"A type with name {self.name} was already registered.")
        context.typedefs[self.name] = self

        self.library = library

    def __deepcopy__(self, memo):
        return self

    def __getattr__(self, name: str) -> Expression:
        for member in self.members:
            if member == name: return Expression(f"{self.name}.{member}", self)
        raise AttributeError(f"Member {name} does not exist on enum type {self.name}.")
    
class FlagType(TypeSpecifier):
    members: list[str]
    inner_type: Optional[type[Flag]]
    def __init__(self, name: str, members: Union[list[str], type[Flag]], register: bool = True, library: str | None = None):
        self.name = name
        self.category = TypeCategory.FLAG
        if isinstance(members, list):
            self.members = members
            self.inner_type = None
        else:
            self.members = members._member_names_
            self.inner_type = members 
        self.register = register

        context = CompilerContext.instance()
        if self.name in context.typedefs:
            raise Exception(f"A type with name {self.name} was already registered.")
        context.typedefs[self.name] = self

        self.library = library

    def __deepcopy__(self, memo):
        return self

    def __getattr__(self, name: str) -> Expression:
        all_members: list[str] = []

        ## 'full' flag with whole members
        if name in self.members:
            return Expression(f"{self.name}.{name}", self)
        
        for character in name:
            found = False
            for member in self.members:
                if member == character: 
                    all_members.append(member)
                    found = True
                    break
            if not found:
                raise AttributeError(f"Member {character} does not exist on flag type {self.name}.")

        return Expression(f"{self.name}.{name}", self)
    
class TupleType(TypeSpecifier):
    members: list[TypeSpecifier]
    def __init__(self, name: str, category: TypeCategory, members: list[TypeSpecifier]):
        self.name = name
        self.category = category
        self.members = members

StateTypeT = EnumType("StateType", StateType, register = False)
MoveTypeT = EnumType("MoveType", MoveType, register = False)
PhysicsTypeT = EnumType("PhysicsType", PhysicsType, register = False)

HitTypeF = FlagType("HitType", HitType, register = False)
HitAttrF = FlagType("HitAttr", HitAttr, register = False)

TransTypeT = EnumType("TransType", TransType, register = False)
AssertTypeT = EnumType("AssertType", AssertType, register = False)
WaveTypeT = EnumType("WaveType", WaveType, register = False)
HelperTypeT = EnumType("HelperType", HelperType, register = False)
TeamTypeT = EnumType("TeamType", TeamType, register = False)
HitAnimTypeT = EnumType("HitAnimType", HitAnimType, register = False)
AttackTypeT = EnumType("AttackType", AttackType, register = False)
PriorityTypeT = EnumType("PriorityType", PriorityType, register = False)
PosTypeT = EnumType("PosType", PosType, register = False)
SpaceTypeT = EnumType("SpaceType", SpaceType, register = False)
TeamModeTypeT = EnumType("TeamModeType", TeamModeType, register = False)

## TODO: +/- won't work.
HitFlagTypeF = FlagType("HitFlag", HitFlagType, register = False)
GuardFlagTypeF = FlagType("GuardFlag", GuardFlagType, register = False)

ColorType = TupleType("ColorType", TypeCategory.TUPLE, [IntType, IntType, IntType])
ColorMultType = TupleType("ColorMultType", TypeCategory.TUPLE, [FloatType, FloatType, FloatType])
FloatPairType = TupleType("FloatPairType", TypeCategory.TUPLE, [FloatType, FloatType])
FloatPosType = TupleType("FloatPosType", TypeCategory.TUPLE, [FloatType, FloatType, PosTypeT])
IntPairType = TupleType("IntPairType", TypeCategory.TUPLE, [IntType, IntType])
BoolPairType = TupleType("BoolPairType", TypeCategory.TUPLE, [BoolType, BoolType])
WaveTupleType = TupleType("WaveTupleType", TypeCategory.TUPLE, [IntType, FloatType, FloatType, FloatType])
HitStringType = TupleType("HitStringType", TypeCategory.TUPLE, [HitTypeF, HitAttrF])
PriorityPairType = TupleType("PriorityPairType", TypeCategory.TUPLE, [IntType, PriorityTypeT])
SoundPairType = TupleType("SoundPairType", TypeCategory.TUPLE, [SoundType, IntType])
PeriodicColorType = TupleType("PeriodicColorType", TypeCategory.TUPLE, [IntType, IntType, IntType, IntType])

__all__ = [
    "StructureMember", "StructureType", "EnumType", "FlagType", "TupleType",
    "StateTypeT", "MoveTypeT", "PhysicsTypeT",
    "ColorType", "ColorMultType", "TransTypeT", "AssertTypeT", "FloatPairType",
    "WaveTypeT", "HelperTypeT", "HitFlagTypeF", "GuardFlagTypeF", "TeamTypeT", "HitAnimTypeT",
    "AttackTypeT", "PriorityTypeT", "PosTypeT", "FloatPosType", "IntPairType",
    "WaveTupleType", "HitTypeF", "HitAttrF", "HitStringType", "PriorityPairType",
    "SoundPairType", "PeriodicColorType", "BoolPairType", "TeamModeTypeT",
    "SpaceTypeT"
]


## this module provides a wrapper around animations in MUGEN.
## MDK will compile each animation to an AIR definition and then merge these definitions with the provided AIR file, if one is provided.

from __future__ import annotations

from typing import Callable
from enum import Flag, Enum
from copy import deepcopy

from mdk.types.context import CompilerContext

class AnimationFlip(Flag):
    NONE = 0
    VERTICAL = 1
    HORIZONTAL = 2
    BOTH = 3

def get_flip(flip: AnimationFlip):
    if flip == AnimationFlip.BOTH: return "VH"
    if flip == AnimationFlip.HORIZONTAL: return "H"
    if flip == AnimationFlip.VERTICAL: return "V"
    return ""

class Clsn:
    _xmin: int
    _ymin: int
    _xmax: int
    _ymax: int
    _default: bool

    def __init__(self, xmin: int, ymin: int, xmax: int, ymax: int, default: bool = False):
        self._xmin = xmin
        self._ymin = ymin
        self._xmax = xmax
        self._ymax = ymax
        self._default = default

class Frame:
    _group: int
    _index: int
    _length: int = -1
    _offset: tuple[int, int] = (0, 0)
    _flip: AnimationFlip = AnimationFlip.NONE
    _trans: str | None = None
    _scale: tuple[float, float] = (1, 1)
    _rotate: int = 0
    _loopstart: bool = False
    _clsn1: list[Clsn] = []
    _clsn2: list[Clsn] = []
    _last_clsn: Clsn | None = None

    def __init__(self, 
                 group: int, index: int, length: int = -1, 
                 offset: tuple[int, int] = (0, 0), flip: AnimationFlip = AnimationFlip.NONE, trans: str | None = None,
                 scale: tuple[float, float] = (1, 1), rotate: int = 0):
        self._group = group
        self._index = index
        self._length = length
        self._offset = offset
        self._flip = flip
        self._trans = trans
        self._scale = scale
        self._rotate = rotate
        self._loopstart = False
        self._clsn1 = []
        self._clsn2 = []
        self._last_clsn = None

    def compile(self):
        """Compiles the current Frame, returning a single line of an animation."""
        result = ""
        
        ## Clsn1Default
        clsns = [c for c in self._clsn1 if c._default]
        if len(clsns) != 0: result = f"{result}Clsn1Default: {len(clsns)}\n"
        for index in range(len(clsns)):
            clsn = clsns[index]
            result = f"{result}  Clsn1[{index}] = {clsn._xmin}, {clsn._ymin}, {clsn._xmax}, {clsn._ymax}\n"
        ## Clsn2Default
        clsns = [c for c in self._clsn2 if c._default]
        if len(clsns) != 0: result = f"{result}Clsn2Default: {len(clsns)}\n"
        for index in range(len(clsns)):
            clsn = clsns[index]
            result = f"{result}  Clsn2[{index}] = {clsn._xmin}, {clsn._ymin}, {clsn._xmax}, {clsn._ymax}\n"
        ## Clsn1
        clsns = [c for c in self._clsn1 if not c._default]
        if len(clsns) != 0: result = f"{result}Clsn1: {len(clsns)}\n"
        for index in range(len(clsns)):
            clsn = clsns[index]
            result = f"{result}  Clsn1[{index}] = {clsn._xmin}, {clsn._ymin}, {clsn._xmax}, {clsn._ymax}\n"
        ## Clsn2
        clsns = [c for c in self._clsn2 if not c._default]
        if len(clsns) != 0: result = f"{result}Clsn2: {len(clsns)}\n"
        for index in range(len(clsns)):
            clsn = clsns[index]
            result = f"{result}  Clsn2[{index}] = {clsn._xmin}, {clsn._ymin}, {clsn._xmax}, {clsn._ymax}\n"

        ## Loopstart
        if self._loopstart:
            result = f"{result}Loopstart\n"

        result = f"{result}{self._group}, {self._index}, {self._offset[0]}, {self._offset[1]}, {self._length}"

        ## early exit if no extra params are set.
        if self._flip == AnimationFlip.NONE and self._trans == None and self._scale[0] == 1 and self._scale[1] == 1 and self._rotate == 0:
            return result
        result += f", {get_flip(self._flip)}"
        result += f", {self._trans if self._trans != None else ''}"
        result += f", {self._scale[0]}, {self._scale[1]}"
        result += f", {self._rotate}"
        return result
    
    def python(self, indent: int = 0) -> str:
        indents = "    " * (indent + 1)

        result = f"Frame({self._group}, {self._index}, length = {self._length})"

        if self._offset != (0, 0): result += f"\n{indents}.offset({self._offset})"
        if self._flip != AnimationFlip.NONE: result += f"\n{indents}.flip({self._flip})"
        if self._trans != None: result += f"\n{indents}.trans(\"{self._trans}\")"
        if self._scale != (1, 1): result += f"\n{indents}.scale({self._scale})"
        if self._rotate != 0: result += f"\n{indents}.rotate({self._rotate})"
        if self._loopstart: result += f"\n{indents}.loop()"
        
        for clsn in self._clsn1:
            result += f"\n{indents}.clsn1(Clsn({clsn._xmin}, {clsn._ymin}, {clsn._xmax}, {clsn._ymax}))"
            if clsn._default: result += ".default()"

        for clsn in self._clsn2:
            result += f"\n{indents}.clsn2(Clsn({clsn._xmin}, {clsn._ymin}, {clsn._xmax}, {clsn._ymax}))"
            if clsn._default: result += ".default()"

        return result
    
    def seq(self) -> Sequence:
        """Converts this Frame into a Sequence containing the frame."""
        return Sequence([self])
    
    def dup(self) -> Frame:
        """Returns a new Frame which is a duplicate of this Frame."""
        return deepcopy(self)
    
    def clsn1(self, clsn: Clsn) -> Frame:
        """Attaches a Clsn1 (hitbox) to the character."""
        self._clsn1.append(deepcopy(clsn))
        self._last_clsn = self._clsn1[-1]
        return self
    
    def clsn2(self, clsn: Clsn) -> Frame:
        """Attaches a Clsn2 (hurtbox) to the character."""
        self._clsn2.append(deepcopy(clsn))
        self._last_clsn = self._clsn2[-1]
        return self
    
    def default(self) -> Frame:
        """Makes the most recently applied Clsn into a default Clsn."""
        if self._last_clsn == None:
            raise Exception("Can only call `default` if a Clsn has been defined!")
        self._last_clsn._default = True
        return self
    
    def length(self, length: int) -> Frame:
        """Updates the duration of the frame to the provided length."""
        self._length = length
        return self
    
    def flip(self, flip: AnimationFlip) -> Frame:
        """Adds the provided flip type(s) to the frame."""
        self._flip = self._flip | flip
        return self
    
    def rotate(self, rotate: int) -> Frame:
        """Applies the provided angle of rotation to the frame."""
        self._rotate = rotate
        return self
        
    def offset(self, offset: tuple[int, int]) -> Frame:
        """Applies the provided offset to the frame."""
        self._offset = offset
        return self
    
    def translate(self, offset: tuple[int, int]) -> Frame:
        """Translates the frame, adding the provided offset to its internal offset."""
        self._offset = (self._offset[0] + offset[0], self._offset[1] + offset[1])
        return self
    
    def scale(self, scale: tuple[float, float]) -> Frame:
        """Applies the provided scale to the frame."""
        self._scale = scale
        return self
    
    def loop(self) -> Frame:
        """Indicates that the sequence's loop-time should begin on this frame."""
        self._loopstart = True
        return self
    
class SequenceModifier:
    """A special class used to modify properties across a range of Frames."""
    _frames: list[Frame]
    _prop: str
    _slice: slice
    
    def __init__(self, frames: list[Frame], prop: str, sl: slice | None = None):
        self._frames = deepcopy(frames)
        self._prop = prop
        self._slice = sl if sl != None else slice(0, len(frames))

    def python(self, name: str | None = None):
        return self.seq().python(name)

    def seq(self):
        """Converts this modified Sequence into a real Sequence object."""
        return Sequence(self._frames)
    
    def set(self, val: int | Enum, filter: Callable[[Frame], bool] | None = None) -> SequenceModifier:
        """Sets the provided value to the target property on all frames (or frames which pass the provided filter function)."""
        for frame in self._frames[self._slice]:
            if filter == None or filter(frame):
                setattr(frame, self._prop, val)
        return self
    
    def add(self, val: int | Enum, filter: Callable[[Frame], bool] | None = None) -> SequenceModifier:
        """Adds the provided value to the target property on all frames (or frames which pass the provided filter function)."""
        for frame in self._frames[self._slice]:
            if filter == None or filter(frame):
                setattr(frame, self._prop, getattr(frame, self._prop) + val)
        return self
    
    def mul(self, val: int | Enum, filter: Callable[[Frame], bool] | None = None) -> SequenceModifier:
        """Multiplies the provided value to the target property on all frames (or frames which pass the provided filter function)."""
        for frame in self._frames[self._slice]:
            if filter == None or filter(frame):
                setattr(frame, self._prop, getattr(frame, self._prop) * val)
        return self
    
    def transform(self, transformer: Callable[[Frame, int], int]) -> SequenceModifier:
        """Uses the provided transformer function to transform the property on all frames."""
        for frame in self._frames[self._slice]:
            setattr(frame, self._prop, transformer(frame, getattr(frame, self._prop)))
        return self
    
    def extend(self, frames: Frame | list[Frame] | Sequence | SequenceModifier | FrameSequenceModifier | TupleSequenceModifier) -> Sequence:
        """Adds the provided frame or frames to this Sequence and returns a new Sequence containing those frames."""
        return self.seq().extend(frames)
    
    def __getitem__(self, key: int | slice) -> SequenceModifier:
        if type(key) == slice:
            self._slice = key
            return self
        else:
            self._slice = slice(key, key + 1)
            return self
    
    ### below this is all the Sequence properties re-applied here. this is for convenience to avoid needing to call `seq()` every time.
    def compile(self):
        return self.seq().compile()

    @property
    def frames(self):
        """Returns a modifiable Sequence which can apply generic transformations across all contained frames."""
        return FrameSequenceModifier(self._frames, self._slice)
        
    @property
    def group(self):
        """Returns a modifiable Sequence which can apply transformations to the `group` property of all contained frames."""
        return SequenceModifier(self._frames, "_group", self._slice)
    
    @property
    def index(self):
        """Returns a modifiable Sequence which can apply transformations to the `index` property of all contained frames."""
        return SequenceModifier(self._frames, "_index", self._slice)
    
    @property
    def length(self):
        """Returns a modifiable Sequence which can apply transformations to the `length` property of all contained frames."""
        return SequenceModifier(self._frames, "_length", self._slice)
    
    @property
    def rotation(self):
        """Returns a modifiable Sequence which can apply transformations to the `rotate` property of all contained frames."""
        return SequenceModifier(self._frames, "_rotate", self._slice)
    
    @property
    def offset(self):
        """Returns a modifiable Sequence which can apply transformations to the `offset` property of all contained frames."""
        return TupleSequenceModifier(self._frames, "_offset", self._slice)
    
    @property
    def scale(self):
        """Returns a modifiable Sequence which can apply transformations to the `scale` property of all contained frames."""
        return TupleSequenceModifier(self._frames, "_scale", self._slice)
    
    @property
    def flip(self):
        """Returns a modifiable Sequence which can apply transformations to the `flip` property of all contained frames."""
        return SequenceModifier(self._frames, "_flip", self._slice)
    
class TupleSequenceModifier(SequenceModifier):
    """A custom SequenceModifier used to apply transformations to the tuple fields, offset and scale."""

    def set(self, val: tuple[int, int], filter: Callable[[Frame], bool] | None = None) -> TupleSequenceModifier: # type: ignore
        """Sets the provided value to the target property on all frames (or frames which pass the provided filter function)."""
        for frame in self._frames[self._slice]:
            if filter == None or filter(frame):
                setattr(frame, self._prop, val)
        return self
    
    def add(self, val: tuple[int, int], filter: Callable[[Frame], bool] | None = None) -> TupleSequenceModifier: # type: ignore
        """Adds the provided value to the target property on all frames (or frames which pass the provided filter function)."""
        for frame in self._frames[self._slice]:
            if filter == None or filter(frame):
                current: tuple[int, int] = getattr(frame, self._prop)
                setattr(frame, self._prop, (current[0] + val[0], current[1] + val[1]))
        return self
    
    def mul(self, val: tuple[int, int], filter: Callable[[Frame], bool] | None = None) -> TupleSequenceModifier: # type: ignore
        """Multiplies the provided value to the target property on all frames (or frames which pass the provided filter function)."""
        for frame in self._frames[self._slice]:
            if filter == None or filter(frame):
                current: tuple[int, int] = getattr(frame, self._prop)
                setattr(frame, self._prop, (current[0] * val[0], current[1] * val[1]))
        return self
    
    def transform(self, transformer: Callable[[Frame, tuple[int, int]], tuple[int, int]]) -> TupleSequenceModifier: # type: ignore
        """Uses the provided transformer function to transform the property on all frames."""
        for frame in self._frames[self._slice]:
            setattr(frame, self._prop, transformer(frame, getattr(frame, self._prop)))
        return self
    
class FrameSequenceModifier(SequenceModifier):
    """A custom SequenceModifier used to apply transformations across the frame sequence, instead of to frame properties."""

    def __init__(self, frames: list[Frame], sl: slice | None = None):
        self._frames = deepcopy(frames)
        self._slice = sl if sl != None else slice(0, len(frames))

    def set(self, val: int | Enum, filter: Callable[[Frame], bool] | None = None) -> SequenceModifier:
        raise Exception("SequenceModifier `set` cannnot be used for the `frames` modifiable sequence.")
    def add(self, val: int | Enum, filter: Callable[[Frame], bool] | None = None) -> SequenceModifier:
        raise Exception("SequenceModifier `add` cannnot be used for the `frames` modifiable sequence.")
    def mul(self, val: int | Enum, filter: Callable[[Frame], bool] | None = None) -> SequenceModifier:
        raise Exception("SequenceModifier `mul` cannnot be used for the `frames` modifiable sequence.")
    
    def reverse(self) -> FrameSequenceModifier:
        """Reverses the Frames contained in this modifiable Sequence.
        This function does not consider the current slice, and will reverse ALL frames in the sequence."""
        self._frames = list(reversed(self._frames))
        return self
    
    def transform(self, transformer: Callable[[Frame], Frame]) -> FrameSequenceModifier: # type: ignore
        """Applies the provided transformation function to all Frames in the sequence, updating the Frames with the returned value."""
        newframes: list[Frame] = []
        for frame in self._frames:
            if frame in self._frames[self._slice]:
                newframes.append(transformer(frame))
            else:
                newframes.append(frame)
        self._frames = newframes
        return self
    
    def clsn1(self, clsn: Clsn, filter: Callable[[Frame], bool] | None = None) -> FrameSequenceModifier:
        """Applies the provided Clsn1 box to all frames which match the provided filter."""
        for frame in self._frames[self._slice]:
            if filter == None or filter(frame):
                frame.clsn1(clsn)
        return self
    
    def clsn2(self, clsn: Clsn, filter: Callable[[Frame], bool] | None = None) -> FrameSequenceModifier:
        """Applies the provided Clsn1 box to all frames which match the provided filter."""
        for frame in self._frames[self._slice]:
            if filter == None or filter(frame):
                frame.clsn2(clsn)
        return self
    
    def default(self, filter: Callable[[Frame], bool] | None = None) -> FrameSequenceModifier:
        """Makes the most recently applied Clsn box default for all frames which match the provided filter."""
        for frame in self._frames[self._slice]:
            if filter == None or filter(frame):
                frame.default()
        return self

class Sequence:
    """Represents a sequence of animation elements (or Frames).
    A Sequence is immutable, but there are multiple utility methods to create a modified copy of a Sequence."""
    _frames: list[Frame]

    def __init__(self, frames: list[Frame]):
        self._frames = frames

    def compile(self):
        """Compiles the frames contained in this Sequence, returning a collection of animation elements in AIR format."""
        result = ""
        for frame in self._frames:
            result += frame.compile() + "\n"
        return result
    
    def python(self, name: str | None = None) -> str:
        if name == None:
            name = "(Anonymous Member)"

        result = f"{name} = Sequence([\n"

        for frame in self._frames:
            result += f"        {frame.python(2)},\n"

        result += "])"
        
        return result
    
    def extend(self, frames: Frame | list[Frame] | Sequence | SequenceModifier | FrameSequenceModifier | TupleSequenceModifier) -> Sequence:
        """Adds the provided frame or frames to this Sequence and returns a new Sequence containing those frames."""
        newframes = deepcopy(self._frames)

        if type(frames) == FrameSequenceModifier or type(frames) == TupleSequenceModifier:
            frames = frames.seq()

        if type(frames) == Frame:
            return Sequence(newframes + [frames])
        elif type(frames) == list and all(isinstance(n, Frame) for n in frames):
            return Sequence(newframes + frames)
        elif type(frames) == Sequence:
            return Sequence(newframes + deepcopy(frames._frames))
        elif type(frames) == SequenceModifier:
            return self.extend(frames.seq())
        raise Exception(f"Unexpected input type {type(frames)} to Sequence extension")
    
    def __getitem__(self, key: int | slice) -> Sequence:
        if type(key) == slice:
            if key.step != None:
                raise Exception("Can only handle simple slices in Sequence.")
            return Sequence(deepcopy(self._frames[key.start:key.stop]))
        else:
            return Sequence([deepcopy(self._frames[key])])
    
    @property
    def frames(self):
        """Returns a modifiable Sequence which can apply generic transformations across all contained frames."""
        return FrameSequenceModifier(self._frames)
    
    @property
    def group(self):
        """Returns a modifiable Sequence which can apply transformations to the `group` property of all contained frames."""
        return SequenceModifier(self._frames, "_group")
    
    @property
    def index(self):
        """Returns a modifiable Sequence which can apply transformations to the `index` property of all contained frames."""
        return SequenceModifier(self._frames, "_index")
    
    @property
    def length(self):
        """Returns a modifiable Sequence which can apply transformations to the `length` property of all contained frames."""
        return SequenceModifier(self._frames, "_length")
    
    @property
    def rotation(self):
        """Returns a modifiable Sequence which can apply transformations to the `rotate` property of all contained frames."""
        return SequenceModifier(self._frames, "_rotate")
    
    @property
    def offset(self):
        """Returns a modifiable Sequence which can apply transformations to the `offset` property of all contained frames."""
        return TupleSequenceModifier(self._frames, "_offset")
    
    @property
    def scale(self):
        """Returns a modifiable Sequence which can apply transformations to the `scale` property of all contained frames."""
        return TupleSequenceModifier(self._frames, "_scale")
    
    @property
    def flip(self):
        """Returns a modifiable Sequence which can apply transformations to the `flip` property of all contained frames."""
        return SequenceModifier(self._frames, "_flip")

class Animation:
    """Represents an animation in a character's AIR file.
    Animations are immutable, once an animation has been defined it should not be modified."""
    _frames: Sequence | None
    _id: int | None
    _external: bool

    def __init__(self, frames: Sequence | SequenceModifier | list[Frame] | Frame | None = None, id: int | None = None, external: bool = False):
        if frames != None and external:
            raise Exception("External animation must not provide a frames parameter.")
        
        self._external = False
        
        if frames == None or external:
            self._frames = None
            self._external = True
        elif type(frames) == list and all(isinstance(n, Frame) for n in frames):
            self._frames = Sequence(frames)
        elif type(frames) == Frame:
            self._frames = Sequence([frames])
        elif type(frames) == SequenceModifier or type(frames) == FrameSequenceModifier or type(frames) == TupleSequenceModifier:
            self._frames = frames.seq()
        elif type(frames) == Sequence:
            self._frames = frames
        else:
            raise Exception("frames must be None, a Sequence, or a list of Frame objects.")

        self._id = id

        matching = [x for x in CompilerContext.instance().animations if x._id == self._id and not x._external]
        if self._id != None and not self._external and len(matching) != 0:
            raise Exception(f"An animation with ID {self._id} was already declared!")
        CompilerContext.instance().animations.append(self)

    @property
    def sequence(self):
        """Returns a modifiable Sequence object which can be used to transform and obtain a new Sequence."""
        return deepcopy(self._frames) if self._frames != None else Sequence([])

    def compile(self):
        """Compiles the animation, returning a string in AIR format."""

        ## assign an animation ID based on the IDs which have already been used
        if self._id == None:
            self._id = CompilerContext.instance().get_next_anim_id()
        
        ## build the AIR output from the animation ID and the contained sequence
        result = f"[Begin Action {self._id}]\n"
        result += self._frames.compile() if self._frames != None else ""

        return result
    
    def python(self, name: str | None = None) -> str:
        if name == None:
            name = "(Anonymous Member)"
        
        result = f"{name} = Animation(\n"
        if self._id != None:
            result += f"    id = {self._id},\n"
        result += f"    frames = Sequence([\n"

        if self._frames != None:
            for frame in self._frames._frames:
                result += f"        {frame.python(2)},\n"

        result += "    ])\n"
        result += ")"

        return result
    
    def __str__(self):
        return self.compile()
    
    def __repr__(self):
        return self.compile()
    
__all__ = ["Animation", "Sequence", "Frame", "AnimationFlip"]


from typing import Optional, Callable

from mdk.types.context import CompilerContext
from mdk.types.expressions import Expression, ConvertibleExpression
import mdk.stdlib.triggers as triggers

from mdk.utils.shared import convert

class RedirectTarget:
    def __init__(self, target: str, expr: Optional[ConvertibleExpression] = None):
        self.target = target
        self.expr = convert(expr) if expr != None else None

        ## create redirected trigger expressions.
        ## this is littered with type-check failures, they are safe to ignore.
        self.AiLevel = self._redirect_atom(triggers.AiLevel)
        self.Alive = self._redirect_atom(triggers.Alive)
        self.Anim = self._redirect_atom(triggers.Anim)
        self.AnimElemNo: Callable[..., Expression] = self._redirect_function(triggers.AnimElemNo)  # type: ignore
        self.AnimElemTime: Callable[..., Expression] = self._redirect_function(triggers.AnimElemTime)  # type: ignore
        self.AnimExist: Callable[..., Expression] = self._redirect_function(triggers.AnimExist)  # type: ignore
        self.AnimTime = self._redirect_atom(triggers.AnimTime)
        self.AuthorName = self._redirect_atom(triggers.AuthorName)
        self.BackEdgeBodyDist = self._redirect_atom(triggers.BackEdgeBodyDist)
        self.BackEdgeDist = self._redirect_atom(triggers.BackEdgeDist)
        self.CanRecover = self._redirect_atom(triggers.CanRecover)
        self.Command = self._redirect_atom(triggers.Command)
        self.Cond: Callable[..., Expression] = self._redirect_function(triggers.Cond)
        self.Const: Callable[..., Expression] = self._redirect_function(triggers.Const)  # type: ignore
        self.Ctrl = self._redirect_atom(triggers.Ctrl)
        self.DrawGame = self._redirect_atom(triggers.DrawGame)
        self.Facing = self._redirect_atom(triggers.Facing)
        self.FrontEdgeBodyDist = self._redirect_atom(triggers.FrontEdgeBodyDist)
        self.FrontEdgeDist = self._redirect_atom(triggers.FrontEdgeDist)
        self.GetHitVar: Callable[..., Expression] = self._redirect_function(triggers.GetHitVar)
        self.HitCount = self._redirect_atom(triggers.HitCount)
        self.HitDefAttr = self._redirect_atom(triggers.HitDefAttr)
        self.HitFall = self._redirect_atom(triggers.HitFall)
        self.HitOver = self._redirect_atom(triggers.HitOver)
        self.HitPauseTime = self._redirect_atom(triggers.HitPauseTime)
        self.HitShakeOver = self._redirect_atom(triggers.HitShakeOver)
        self.HitVel = self._redirect_position(triggers.HitVel)
        self.ID = self._redirect_atom(triggers.ID)
        self.InGuardDist = self._redirect_atom(triggers.InGuardDist)
        self.IsHelper: Callable[..., Expression] = self._redirect_function(triggers.IsHelper) # type: ignore
        self.IsHomeTeam = self._redirect_atom(triggers.IsHomeTeam)
        self.Life = self._redirect_atom(triggers.Life)
        self.LifeMax = self._redirect_atom(triggers.LifeMax)
        self.Lose = self._redirect_atom(triggers.Lose)
        self.MoveContact = self._redirect_atom(triggers.MoveContact)
        self.MoveGuarded = self._redirect_atom(triggers.MoveGuarded)
        self.MoveHit = self._redirect_atom(triggers.MoveHit)
        self.MoveReversed = self._redirect_atom(triggers.MoveReversed)
        self.MoveType = self._redirect_atom(triggers.MoveType)
        self.Name = self._redirect_atom(triggers.Name)
        self.NumEnemy = self._redirect_atom(triggers.NumEnemy)
        self.NumExplod: Callable[..., Expression] = self._redirect_function(triggers.NumExplod) # type: ignore
        self.NumHelper: Callable[..., Expression] = self._redirect_function(triggers.NumHelper) # type: ignore
        self.NumPartner = self._redirect_atom(triggers.NumPartner)
        self.NumProj = self._redirect_atom(triggers.NumProj)
        self.NumProjID: Callable[..., Expression] = self._redirect_function(triggers.NumProjID)  # type: ignore
        self.NumTarget: Callable[..., Expression] = self._redirect_function(triggers.NumTarget) # type: ignore
        self.P1Name = self._redirect_atom(triggers.P1Name)
        self.P2BodyDist = self._redirect_position(triggers.P2BodyDist)
        self.P2Dist = self._redirect_position(triggers.P2Dist)
        self.P2Life = self._redirect_atom(triggers.P2Life)
        self.P2MoveType = self._redirect_atom(triggers.P2MoveType)
        self.P2Name = self._redirect_atom(triggers.P2Name)
        self.P2StateNo = self._redirect_atom(triggers.P2StateNo)
        self.P2StateType = self._redirect_atom(triggers.P2StateType)
        self.P3Name = self._redirect_atom(triggers.P3Name)
        self.P4Name = self._redirect_atom(triggers.P4Name)
        self.PalNo = self._redirect_atom(triggers.PalNo)
        self.ParentDist = self._redirect_position(triggers.ParentDist)
        self.Pos = self._redirect_position(triggers.Pos)
        self.Power = self._redirect_atom(triggers.Power)
        self.PowerMax = self._redirect_atom(triggers.PowerMax)
        self.PrevStateNo = self._redirect_atom(triggers.PrevStateNo)
        self.ProjCancelTime: Callable[..., Expression] = self._redirect_function(triggers.ProjCancelTime)  # type: ignore
        self.ProjContactTime: Callable[..., Expression] = self._redirect_function(triggers.ProjContactTime)  # type: ignore
        self.ProjGuardedTime: Callable[..., Expression] = self._redirect_function(triggers.ProjGuardedTime)  # type: ignore
        self.ProjHitTime: Callable[..., Expression] = self._redirect_function(triggers.ProjHitTime)  # type: ignore
        self.RootDist = self._redirect_position(triggers.RootDist)
        self.ScreenPos = self._redirect_position(triggers.ScreenPos)
        self.SelfAnimExist: Callable[..., Expression] = self._redirect_function(triggers.SelfAnimExist)  # type: ignore
        self.StateNo = self._redirect_atom(triggers.StateNo)
        self.StateType = self._redirect_atom(triggers.StateType)
        self.StageVar = self._redirect_function(triggers.StageVar)  # type: ignore
        self.TeamMode = self._redirect_atom(triggers.TeamMode)
        self.TeamSide = self._redirect_atom(triggers.TeamSide)
        self.Time = self._redirect_atom(triggers.Time)
        self.Vel = self._redirect_position(triggers.Vel)
        self.Win = self._redirect_atom(triggers.Win)
        self.WinKO = self._redirect_atom(triggers.WinKO)
        self.WinPerfect = self._redirect_atom(triggers.WinPerfect)
        self.WinTime = self._redirect_atom(triggers.WinTime)

    def _redirect_atom(self, expr: Expression) -> Expression:
        return expr.make_expression(f"{self.__repr__()},{expr.exprn}")
    
    def _redirect_position(self, expr: triggers.PositionExpression) -> triggers.PositionExpression:
        return triggers.PositionExpression(f"{self.__repr__()},{expr._name}", expr.x.type)

    def _redirect_function(self, fn: Callable[..., Expression]) -> Callable[..., Expression]:
        def _call(*args, **kwargs) -> Expression:
            result = fn(*args, **kwargs)
            return result.make_expression(f"{self.__repr__()},{result.exprn}")
        return _call

    def __repr__(self):
        if isinstance(self.expr, Expression):
            return f"{self.target}({self.expr.__str__()})"
        return self.target
    
    ## redirects need to be able to access their (global) variables.
    ## it's not important to check scope/ownership here, MTL will take care of it.
    def __getattr__(self, name: str) -> Expression:
        ctx = CompilerContext.instance()
        if (var := next(filter(lambda gv: gv.name == name, ctx.globals), None)) == None:
            raise Exception(f"No variable exists with name {name} for redirect {self.target}.")
        return Expression(f"{self.__repr__()},{name}", var.type)

def RedirectTargetBuilder(target: str) -> Callable[[Optional[ConvertibleExpression]], RedirectTarget]:
    def _redirect(id: Optional[ConvertibleExpression] = None) -> RedirectTarget:
        return RedirectTarget(target, id)
    return _redirect

parent = RedirectTarget("parent")
root = RedirectTarget("root")
partner = RedirectTarget("partner")
helper = RedirectTarget("helper")
target = RedirectTarget("target")
enemy = RedirectTarget("enemy")
enemynear = RedirectTarget("enemynear")

helperID = RedirectTargetBuilder("helper")
targetID = RedirectTargetBuilder("target")
enemyID = RedirectTargetBuilder("enemy")
enemynearID = RedirectTargetBuilder("enemynear")
playerID = RedirectTargetBuilder("playerID")

__all__ = ["parent", "root", "partner", "helper", "target", "enemy", "enemynear", "helperID", "targetID", "enemyID", "enemynearID", "playerID"]